/*
 * jTPCCTerminal - Terminal emulator code for jTPCC (transactions)
 *
 * Copyright (C) 2003, Raul Barbosa <rbarbosa@student.dei.uc.pt>
 *
 * jTPCC is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

import java.io.*;
import java.sql.*;
import java.util.*;
import java.text.*;
import javax.swing.*;

public class jTPCCTerminal implements jTPCCConfig, Runnable
{
    private String terminalName;
    private Connection conn = null;
    private Statement stmt = null;
    private ResultSet rs = null;
    private long terminalWarehouseID, terminalDistrictID;
    private int paymentWeight, orderStatusWeight, deliveryWeight, stockLevelWeight;
    private JOutputArea terminalOutputArea, errorOutputArea;
    private boolean debugMessages;
    private jTPCC parent;

    private Random random;
    private long transactionCount = 1, numTransactions, numWarehouses, newOrderCounter;
    private StringBuffer query = null;
    private int result = 0;
    private boolean stopRunningSignal = false;

    public jTPCCTerminal(String terminalName, long terminalWarehouseID, long terminalDistrictID, Connection conn, long numTransactions, JOutputArea terminalOutputArea, JOutputArea errorOutputArea, boolean debugMessages, int paymentWeight, int orderStatusWeight, int deliveryWeight, int stockLevelWeight, int numWarehouses, jTPCC parent) throws SQLException
    {
        this.terminalName = terminalName;
        this.conn = conn;
        this.stmt = conn.createStatement();
        this.stmt.setMaxRows(200);
        this.stmt.setFetchSize(100);
        this.terminalWarehouseID = terminalWarehouseID;
        this.terminalDistrictID = terminalDistrictID;
        this.terminalOutputArea = terminalOutputArea;
        this.errorOutputArea = errorOutputArea;
        this.debugMessages = debugMessages;
        this.parent = parent;
        this.random = new Random(System.currentTimeMillis() * conn.hashCode());
        this.numTransactions = numTransactions;
        this.paymentWeight = paymentWeight;
        this.orderStatusWeight = orderStatusWeight;
        this.deliveryWeight = deliveryWeight;
        this.stockLevelWeight = stockLevelWeight;
        this.numWarehouses = numWarehouses;
        this.newOrderCounter = 0;
        terminalMessage("Terminal \'" + terminalName + "\' has WarehouseID=" + terminalWarehouseID + " and DistrictID=" + terminalDistrictID + ".");
    }

    public void run()
    {
        executeTransactions(numTransactions);
        try
        {
            printMessage("Closing statement and connection...");
            stmt.close();
            conn.close();
        }
        catch(Exception e)
        {
            printMessage("An error occurred!");
            logException(e);
        }

        printMessage("Terminal \'" + terminalName + "\' finished after " + (transactionCount-1) + " transaction(s).");

        parent.signalTerminalEnded(this, newOrderCounter);
    }

    public void stopRunningWhenPossible()
    {
        stopRunningSignal = true;
        printMessage("Terminal received stop signal!");
        printMessage("Finishing current transaction before exit...");
    }

    private void executeTransactions(long numTransactions)
    {
        boolean stopRunning = false;

        if(numTransactions != -1)
            printMessage("Executing " + numTransactions + " transactions...");
        else
            printMessage("Executing for a limited time...");

        for(int i = 0; (i < numTransactions || numTransactions == -1) && !stopRunning; i++)
        {
            long transactionType = randomNumber(1, 100);
            int skippedDeliveries = 0, newOrder = 0;
            String transactionTypeName;


            System.gc();
            long transactionStart = System.currentTimeMillis();

            if(transactionType <= paymentWeight)
            {
                executeTransaction(PAYMENT);
                transactionTypeName = "Payment";
            }
            else if(transactionType <= paymentWeight + stockLevelWeight)
            {
                executeTransaction(STOCK_LEVEL);
                transactionTypeName = "Stock-Level";
            }
            else if(transactionType <= paymentWeight + stockLevelWeight + orderStatusWeight)
            {
                executeTransaction(ORDER_STATUS);
                transactionTypeName = "Order-Status";
            }
            else if(transactionType <= paymentWeight + stockLevelWeight + orderStatusWeight + deliveryWeight)
            {
                skippedDeliveries = executeTransaction(DELIVERY);
                transactionTypeName = "Delivery";
            }
            else
            {
                executeTransaction(NEW_ORDER);
                transactionTypeName = "New-Order";
                newOrderCounter++;
                newOrder = 1;
            }

            System.gc();
            long transactionEnd = System.currentTimeMillis();


            if(!transactionTypeName.equals("Delivery"))
            {
                parent.signalTerminalEndedTransaction(this.terminalName, transactionTypeName, transactionEnd - transactionStart, null, newOrder);
            }
            else
            {
                parent.signalTerminalEndedTransaction(this.terminalName, transactionTypeName, transactionEnd - transactionStart, (skippedDeliveries == 0 ? "None" : "" + skippedDeliveries + " delivery(ies) skipped."), newOrder);
            }


            if(stopRunningSignal) stopRunning = true;
        }
    }

    private int executeTransaction(int transaction)
    {
        int result = 0;

        switch(transaction)
        {
            case NEW_ORDER:
                long districtID = randomNumber(1, 10);
                long customerID = nonUniformRandom(1023, 1, 3000);

                int numItems = (int)randomNumber(5, 15);
                long[] itemIDs = new long[numItems];
                long[] supplierWarehouseIDs = new long[numItems];
                long[] orderQuantities = new long[numItems];
                int allLocal = 1;
                for(int i = 0; i < numItems; i++)
                {
                    itemIDs[i] = nonUniformRandom(8191, 1, 100000);
                    if(randomNumber(1, 100) > 1)
                    {
                        supplierWarehouseIDs[i] = terminalWarehouseID;
                    }
                    else
                    {
                        do
                        {
                            supplierWarehouseIDs[i] = randomNumber(1, numWarehouses);
                        }
                        while(supplierWarehouseIDs[i] == terminalWarehouseID && numWarehouses > 1);
                        allLocal = 0;
                    }
                    orderQuantities[i] = randomNumber(1, 10);
                }
                if(randomNumber(1, 100) == 1)
                    itemIDs[numItems-1] = -12345;

                terminalMessage("\nStarting transaction #" + transactionCount + " (New-Order)...");
                newOrderTransaction(terminalWarehouseID, districtID, customerID, numItems, allLocal, itemIDs, supplierWarehouseIDs, orderQuantities);
                break;


            case PAYMENT:
                districtID = randomNumber(1, 10);

                long x = randomNumber(1, 100);
                long customerDistrictID;
                long customerWarehouseID;
                if(x <= 85)
                {
                    customerDistrictID = districtID;
                    customerWarehouseID = terminalWarehouseID;
                }
                else
                {
                    customerDistrictID = randomNumber(1, 10);
                    do
                    {
                        customerWarehouseID = randomNumber(1, numWarehouses);
                    }
                    while(customerWarehouseID == terminalWarehouseID && numWarehouses > 1);
                }

                long y = randomNumber(1, 100);
                boolean customerByName;
                String customerLastName = null;
                customerID = -1;
                if(y <= 60)
                {
                    customerByName = true;
                    customerLastName = lastName((int)nonUniformRandom(255, 0, 999));
                }
                else
                {
                    customerByName = false;
                    customerID = nonUniformRandom(1023, 1, 3000);
                }

                double paymentAmount = randomNumber(100, 500000)/100.0;

                terminalMessage("\nStarting transaction #" + transactionCount + " (Payment)...");
                paymentTransaction(terminalWarehouseID, customerWarehouseID, paymentAmount, districtID, customerDistrictID, customerID, customerLastName, customerByName);
                break;


            case STOCK_LEVEL:
                long threshold = randomNumber(10, 20);

                terminalMessage("\nStarting transaction #" + transactionCount + " (Stock-Level)...");
                stockLevelTransaction(terminalWarehouseID, terminalDistrictID, threshold);
                break;


            case ORDER_STATUS:
                districtID = randomNumber(1, 10);

                y = randomNumber(1, 100);
                customerLastName = null;
                customerID = -1;
                if(y <= 60)
                {
                    customerByName = true;
                    customerLastName = lastName((int)nonUniformRandom(255, 0, 999));
                }
                else
                {
                    customerByName = false;
                    customerID = nonUniformRandom(1023, 1, 3000);
                }

                terminalMessage("\nStarting transaction #" + transactionCount + " (Order-Status)...");
                orderStatusTransaction(terminalWarehouseID, districtID, customerID, customerLastName, customerByName);
                break;


            case DELIVERY:
                long orderCarrierID = randomNumber(1, 10);

                terminalMessage("\nStarting transaction #" + transactionCount + " (Delivery)...");
                result = deliveryTransaction(terminalWarehouseID, orderCarrierID);
                break;


            default:
                error("EMPTY-TYPE");
                break;
        }
        transactionCount++;

        return result;
    }


    private int deliveryTransaction(long w_id, long o_carrier_id)
    {
        long d_id, no_o_id, c_id;
        double ol_total;
        long[] orderIDs;
        int skippedDeliveries = 0;
        boolean newOrderRemoved;


        try
        {
            orderIDs = new long[10];
            for(d_id = 1; d_id <= 10; d_id++)
            {
                do
                {
                    no_o_id = -1;
                    query = new StringBuffer();
                    query.append("SELECT * FROM (");
                    query.append("SELECT no_o_id");
                    query.append(" FROM new_order");
                    query.append(" WHERE no_d_id = ");
                    query.append(d_id);
                    query.append(" AND no_w_id = ");
                    query.append(w_id);
                    query.append(" ORDER BY no_o_id ASC");
                    query.append(") WHERE rownum = 1");
                    printMessage(query.toString());
                    rs = stmt.executeQuery(query.toString());
                    if(rs.next()) no_o_id = rs.getLong("no_o_id");
                    orderIDs[(int)d_id-1] = no_o_id;
                    rs.close();
                    rs = null;


                    newOrderRemoved = false;
                    if(no_o_id != -1)
                    {
                        query = new StringBuffer();
                        query.append("DELETE FROM new_order");
                        query.append(" WHERE no_d_id = ");
                        query.append(d_id);
                        query.append(" AND no_w_id = ");
                        query.append(w_id);
                        query.append(" AND no_o_id = ");
                        query.append(no_o_id);
                        printMessage(query.toString());
                        result = stmt.executeUpdate(query.toString());
                        if(result > 0) newOrderRemoved = true;
                    }
                }
                while(no_o_id != -1 && !newOrderRemoved);


                if(no_o_id != -1)
                {
                    query = new StringBuffer();
                    query.append("SELECT o_c_id");
                    query.append(" FROM oorder");
                    query.append(" WHERE o_id = ");
                    query.append(no_o_id);
                    query.append(" AND o_d_id = ");
                    query.append(d_id);
                    query.append(" AND o_w_id = ");
                    query.append(w_id);
                    printMessage(query.toString());
                    rs = stmt.executeQuery(query.toString());
                    if(!rs.next()) throw new Exception("O_ID=" + no_o_id + " O_D_ID=" + d_id + " O_W_ID=" + w_id + " not found!");
                    c_id = rs.getLong("o_c_id");
                    rs.close();
                    rs = null;


                    query = new StringBuffer();
                    query.append("UPDATE oorder SET o_carrier_id = ");
                    query.append(o_carrier_id);
                    query.append(" WHERE o_id = ");
                    query.append(no_o_id);
                    query.append(" AND o_d_id = ");
                    query.append(d_id);
                    query.append(" AND o_w_id = ");
                    query.append(w_id);
                    printMessage(query.toString());
                    result = stmt.executeUpdate(query.toString());
                    if(result == 0) throw new Exception("O_ID=" + no_o_id + " O_D_ID=" + d_id + " O_W_ID=" + w_id + " not found!");


                    query = new StringBuffer();
                    query.append("UPDATE order_line SET ol_delivery_d = SYSDATE ");
                    query.append(" WHERE ol_o_id = ");
                    query.append(no_o_id);
                    query.append(" AND ol_d_id = ");
                    query.append(d_id);
                    query.append(" AND ol_w_id = ");
                    query.append(w_id);
                    printMessage(query.toString());
                    result = stmt.executeUpdate(query.toString());
                    if(result == 0) throw new Exception("OL_O_ID=" + no_o_id + " OL_D_ID=" + d_id + " OL_W_ID=" + w_id + " not found!");


                    query = new StringBuffer();
                    query.append("SELECT SUM(ol_amount) AS ol_total");
                    query.append(" FROM order_line");
                    query.append(" WHERE ol_o_id = ");
                    query.append(no_o_id);
                    query.append(" AND ol_d_id = ");
                    query.append(d_id);
                    query.append(" AND ol_w_id = ");
                    query.append(w_id);
                    printMessage(query.toString());
                    rs = stmt.executeQuery(query.toString());
                    if(!rs.next()) throw new Exception("OL_O_ID=" + no_o_id + " OL_D_ID=" + d_id + " OL_W_ID=" + w_id + " not found!");
                    ol_total = rs.getDouble("ol_total");
                    rs.close();
                    rs = null;


                    query = new StringBuffer();
                    query.append("UPDATE customer SET c_balance = c_balance + ");
                    query.append(ol_total);
                    query.append(", c_delivery_cnt = c_delivery_cnt + 1");
                    query.append(" WHERE c_id = ");
                    query.append(c_id);
                    query.append(" AND c_d_id = ");
                    query.append(d_id);
                    query.append(" AND c_w_id = ");
                    query.append(w_id);
                    printMessage(query.toString());
                    result = stmt.executeUpdate(query.toString());
                    if(result == 0) throw new Exception("C_ID=" + c_id + " C_W_ID=" + w_id + " C_D_ID=" + d_id + " not found!");
                }
            }


            conn.commit();


            StringBuffer terminalMessage = new StringBuffer();
            terminalMessage.append("\n+---------------------------- DELIVERY ---------------------------+\n");
            terminalMessage.append(" Date: ");
            terminalMessage.append(getCurrentTime());
            terminalMessage.append("\n\n Warehouse: ");
            terminalMessage.append(w_id);
            terminalMessage.append("\n Carrier:   ");
            terminalMessage.append(o_carrier_id);
            terminalMessage.append("\n\n Delivered Orders\n");
            for(int i = 1; i <= 10; i++)
            {
                if(orderIDs[i-1] >= 0)
                {
                    terminalMessage.append("  District ");
                    terminalMessage.append(i < 10 ? " " : "");
                    terminalMessage.append(i);
                    terminalMessage.append(": Order number ");
                    terminalMessage.append(orderIDs[i-1]);
                    terminalMessage.append(" was delivered.\n");
                }
                else
                {
                    terminalMessage.append("  District ");
                    terminalMessage.append(i < 10 ? " " : "");
                    terminalMessage.append(i);
                    terminalMessage.append(": No orders to be delivered.\n");
                    skippedDeliveries++;
                }
            }
            terminalMessage.append("+-----------------------------------------------------------------+\n\n");
            terminalMessage(terminalMessage.toString());
        }
        catch(Exception e)
        {
            error("DELIVERY");
            logException(e);
            try
            {
                terminalMessage("Performing ROLLBACK...");
                conn.rollback();
            }
            catch(Exception e1)
            {
                error("DELIVERY-ROLLBACK");
                logException(e1);
            }
        }

        return skippedDeliveries;
    }


    private void orderStatusTransaction(long w_id, long d_id, long c_id, String c_last, boolean c_by_name)
    {
        long namecnt, o_id = -1, o_carrier_id = -1;
        double c_balance;
        String c_first, c_middle;
        java.sql.Date entdate = null;
        Vector orderLines = new Vector();


        try
        {
            if(c_by_name)
            {
                query = new StringBuffer();
                query.append("SELECT count(c_id) AS namecnt");
                query.append(" FROM customer");
                query.append(" WHERE c_last = \'");
                query.append(c_last);
                query.append("\' AND c_d_id = ");
                query.append(d_id);
                query.append(" AND c_w_id = ");
                query.append(w_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_LAST=" + c_last + " C_D_ID=" + d_id + " C_W_ID=" + w_id + " not found!");
                namecnt = rs.getLong("namecnt");
                rs.close();
                rs = null;


                query = new StringBuffer();
                query.append("SELECT c_balance, c_first, c_middle, c_id");
                query.append(" FROM customer");
                query.append(" WHERE c_last = \'");
                query.append(c_last);
                query.append("\' AND c_d_id = ");
                query.append(d_id);
                query.append(" AND c_w_id = ");
                query.append(w_id);
                query.append(" ORDER BY c_first ASC");
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_LAST=" + c_last + " C_D_ID=" + d_id + " C_W_ID=" + w_id + " not found!");
                if(namecnt%2 == 1) namecnt++;
                for(int i = 1; i < namecnt / 2; i++) rs.next();
                c_id = rs.getLong("c_id");
                c_first = rs.getString("c_first");
                c_middle = rs.getString("c_middle");
                c_balance = rs.getDouble("c_balance");
                rs.close();
                rs = null;
            }
            else
            {
                query = new StringBuffer();
                query.append("SELECT c_balance, c_first, c_middle, c_last");
                query.append(" FROM customer");
                query.append(" WHERE c_id = ");
                query.append(c_id);
                query.append(" AND c_d_id = ");
                query.append(d_id);
                query.append(" AND c_w_id = ");
                query.append(w_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_ID=" + c_id + " C_D_ID=" + d_id + " C_W_ID=" + w_id + " not found!");
                c_last = rs.getString("c_last");
                c_first = rs.getString("c_first");
                c_middle = rs.getString("c_middle");
                c_balance = rs.getDouble("c_balance");
                rs.close();
                rs = null;
            }


            query = new StringBuffer();
            query.append("SELECT * FROM (");
            query.append("SELECT o_id, o_carrier_id, o_entry_d");
            query.append(" FROM oorder");
            query.append(" WHERE o_w_id = ");
            query.append(w_id);
            query.append(" AND o_d_id = ");
            query.append(d_id);
            query.append(" AND o_c_id = ");
            query.append(c_id);
            query.append(" ORDER BY o_id DESC");
            query.append(") WHERE rownum = 1");
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(rs.next())
            {
                o_id = rs.getLong("o_id");
                o_carrier_id = rs.getLong("o_carrier_id");
                entdate = rs.getDate("o_entry_d");
            }
            rs.close();
            rs = null;


            query = new StringBuffer();
            query.append("SELECT ol_i_id, ol_supply_w_id, ol_quantity,");
            query.append(" ol_amount, ol_delivery_d");
            query.append(" FROM order_line");
            query.append(" WHERE ol_o_id = ");
            query.append(o_id);
            query.append(" AND ol_d_id = ");
            query.append(d_id);
            query.append(" AND ol_w_id = ");
            query.append(w_id);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            while(rs.next())
            {
                StringBuffer orderLine = new StringBuffer();
                orderLine.append("[");
                orderLine.append(rs.getLong("ol_supply_w_id"));
                orderLine.append(" - ");
                orderLine.append(rs.getLong("ol_i_id"));
                orderLine.append(" - ");
                orderLine.append(rs.getLong("ol_quantity"));
                orderLine.append(" - ");
                orderLine.append(formattedDouble(rs.getDouble("ol_amount")));
                orderLine.append(" - ");
                if(rs.getDate("ol_delivery_d") != null)
                    orderLine.append(rs.getDate("ol_delivery_d"));
                else
                    orderLine.append("99-99-9999");
                orderLine.append("]");
                orderLines.add(orderLine.toString());
            }
            rs.close();
            rs = null;


            StringBuffer terminalMessage = new StringBuffer();
            terminalMessage.append("\n");
            terminalMessage.append("+-------------------------- ORDER-STATUS -------------------------+\n");
            terminalMessage.append(" Date: ");
            terminalMessage.append(getCurrentTime());
            terminalMessage.append("\n\n Warehouse: ");
            terminalMessage.append(w_id);
            terminalMessage.append("\n District:  ");
            terminalMessage.append(d_id);
            terminalMessage.append("\n\n Customer:  ");
            terminalMessage.append(c_id);
            terminalMessage.append("\n   Name:    ");
            terminalMessage.append(c_first);
            terminalMessage.append(" ");
            terminalMessage.append(c_middle);
            terminalMessage.append(" ");
            terminalMessage.append(c_last);
            terminalMessage.append("\n   Balance: ");
            terminalMessage.append(c_balance);
            terminalMessage.append("\n\n");
            if(o_id == -1)
            {
                terminalMessage.append(" Customer has no orders placed.\n");
            }
            else
            {
                terminalMessage.append(" Order-Number: ");
                terminalMessage.append(o_id);
                terminalMessage.append("\n    Entry-Date: ");
                terminalMessage.append(entdate);
                terminalMessage.append("\n    Carrier-Number: ");
                terminalMessage.append(o_carrier_id);
                terminalMessage.append("\n\n");
                if(orderLines.size() != 0)
                {
                    terminalMessage.append(" [Supply_W - Item_ID - Qty - Amount - Delivery-Date]\n");
                    Enumeration orderLinesEnum = orderLines.elements();
                    while(orderLinesEnum.hasMoreElements())
                    {
                        terminalMessage.append(" ");
                        terminalMessage.append((String)orderLinesEnum.nextElement());
                        terminalMessage.append("\n");
                    }
                }
                else
                {
                    terminalMessage(" This Order has no Order-Lines.\n");
                }
            }
            terminalMessage.append("+-----------------------------------------------------------------+\n\n");
            terminalMessage(terminalMessage.toString());
        }
        catch(Exception e)
        {
            error("ORDER-STATUS");
            logException(e);
        }
    }


    private void newOrderTransaction(long w_id, long d_id, long c_id, int o_ol_cnt, int o_all_local, long[] itemIDs, long[] supplierWarehouseIDs, long[] orderQuantities)
    {
        double c_discount, w_tax, d_tax = 0, i_price;
        long d_next_o_id, o_id = -1, s_quantity;
        String c_last = null, c_credit = null, i_name, i_data, s_data;
        String s_dist_01, s_dist_02, s_dist_03, s_dist_04, s_dist_05;
        String s_dist_06, s_dist_07, s_dist_08, s_dist_09, s_dist_10, ol_dist_info = null;
        double[] itemPrices = new double[o_ol_cnt];
        double[] orderLineAmounts = new double[o_ol_cnt];
        String[] itemNames = new String[o_ol_cnt];
        long[] stockQuantities = new long[o_ol_cnt];
        char[] brandGeneric = new char[o_ol_cnt];
        long ol_supply_w_id, ol_i_id, ol_quantity;
        int s_remote_cnt_increment;
        double ol_amount, total_amount = 0;
        boolean newOrderRowInserted;


        try
        {
            query = new StringBuffer();
            query.append("SELECT c_discount, c_last, c_credit, w_tax");
            query.append(" FROM customer, warehouse");
            query.append(" WHERE w_id = ");
            query.append(w_id);
            query.append(" AND c_w_id = ");
            query.append(w_id);
            query.append(" AND c_d_id = ");
            query.append(d_id);
            query.append(" AND c_id = ");
            query.append(c_id);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(!rs.next()) throw new Exception("W_ID=" + w_id + " C_D_ID=" + d_id + " C_ID=" + c_id + " not found!");
            c_discount = rs.getDouble("c_discount");
            c_last = rs.getString("c_last");
            c_credit = rs.getString("c_credit");
            w_tax = rs.getDouble("w_tax");
            rs.close();
            rs = null;


            newOrderRowInserted = false;
            while(!newOrderRowInserted)
            {
                query = new StringBuffer();
                query.append("SELECT d_next_o_id, d_tax");
                query.append(" FROM district");
                query.append(" WHERE d_id = ");
                query.append(d_id);
                query.append(" AND d_w_id = ");
                query.append(w_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("D_ID=" + d_id + " D_W_ID=" + w_id + " not found!");
                d_next_o_id = rs.getLong("d_next_o_id");
                d_tax = rs.getDouble("d_tax");
                rs.close();
                rs = null;


                o_id = d_next_o_id;


                try
                {
                    query = new StringBuffer();
                    query.append("INSERT INTO NEW_ORDER (no_o_id, no_d_id, no_w_id)");
                    query.append(" VALUES (");
                    query.append(o_id);
                    query.append(", ");
                    query.append(d_id);
                    query.append(", ");
                    query.append(w_id);
                    query.append(")");
                    printMessage(query.toString());
                    stmt.executeUpdate(query.toString());
                    newOrderRowInserted = true;
                }
                catch(SQLException e2)
                {
                    printMessage("The row was already on table new_order. Restarting...");
                }
            }


            query = new StringBuffer();
            query.append("UPDATE district SET d_next_o_id = d_next_o_id + 1");
            query.append(" WHERE d_id = ");
            query.append(d_id);
            query.append(" AND d_w_id = ");
            query.append(w_id);
            printMessage(query.toString());
            result = stmt.executeUpdate(query.toString());
            if(result == 0) throw new Exception("D_ID=" + d_id + " D_W_ID=" + w_id + " not found!");


            query = new StringBuffer();
            query.append("INSERT INTO OORDER (o_id, o_d_id, o_w_id, o_c_id,");
            query.append(" o_entry_d, o_ol_cnt, o_all_local, o_carrier_id)");
            query.append(" VALUES (");
            query.append(o_id);
            query.append(", ");
            query.append(d_id);
            query.append(", ");
            query.append(w_id);
            query.append(", ");
            query.append(c_id);
            query.append(", SYSDATE");
            query.append(", ");
            query.append(o_ol_cnt);
            query.append(", ");
            query.append(o_all_local);
            query.append(", NULL)");
            printMessage(query.toString());
            stmt.executeUpdate(query.toString());


            for(int ol_number = 1; ol_number <= o_ol_cnt; ol_number++)
            {
                ol_supply_w_id = supplierWarehouseIDs[ol_number-1];
                ol_i_id = itemIDs[ol_number-1];
                ol_quantity = orderQuantities[ol_number-1];


                query = new StringBuffer();
                query.append("SELECT i_price, i_name , i_data");
                query.append(" FROM item");
                query.append(" WHERE i_id = ");
                query.append(ol_i_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new IllegalAccessException("I_ID=" + ol_i_id + " not found!");
                i_price = rs.getDouble("i_price");
                i_name = rs.getString("i_name");
                i_data = rs.getString("i_data");
                rs.close();
                rs = null;


                itemPrices[ol_number-1] = i_price;
                itemNames[ol_number-1] = i_name;


                query = new StringBuffer();
                query.append("SELECT s_quantity, s_data,");
                query.append(" s_dist_01, s_dist_02, s_dist_03, s_dist_04, s_dist_05,");
                query.append(" s_dist_06, s_dist_07, s_dist_08, s_dist_09, s_dist_10");
                query.append(" FROM stock");
                query.append(" WHERE s_i_id = ");
                query.append(ol_i_id);
                query.append(" AND s_w_id = ");
                query.append(ol_supply_w_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("I_ID=" + ol_i_id + " not found!");
                s_quantity = rs.getLong("s_quantity");
                s_data = rs.getString("s_data");
                s_dist_01 = rs.getString("s_dist_01");
                s_dist_02 = rs.getString("s_dist_02");
                s_dist_03 = rs.getString("s_dist_03");
                s_dist_04 = rs.getString("s_dist_04");
                s_dist_05 = rs.getString("s_dist_05");
                s_dist_06 = rs.getString("s_dist_06");
                s_dist_07 = rs.getString("s_dist_07");
                s_dist_08 = rs.getString("s_dist_08");
                s_dist_09 = rs.getString("s_dist_09");
                s_dist_10 = rs.getString("s_dist_10");
                rs.close();
                rs = null;


                stockQuantities[ol_number-1] = s_quantity;

                if(s_quantity - ol_quantity >= 10)
                {
                    s_quantity -= ol_quantity;
                }
                else
                {
                    s_quantity += -ol_quantity + 91;
                }

                if(ol_supply_w_id == w_id)
                {
                    s_remote_cnt_increment = 0;
                }
                else
                {
                    s_remote_cnt_increment = 1;
                }

                query = new StringBuffer();
                query.append("UPDATE stock SET s_quantity = ");
                query.append(s_quantity);
                query.append(", s_ytd = s_ytd + ");
                query.append(ol_quantity);
                query.append(", s_remote_cnt = s_remote_cnt + ");
                query.append(s_remote_cnt_increment);
                query.append(" WHERE s_i_id = ");
                query.append(ol_i_id);
                query.append(" AND s_w_id = ");
                query.append(ol_supply_w_id);
                printMessage(query.toString());
                result = stmt.executeUpdate(query.toString());
                if(result == 0) throw new Exception("S_I_ID=" + ol_i_id + " S_W_ID=" + ol_supply_w_id + " not found!");


                ol_amount = ol_quantity * i_price;
                orderLineAmounts[ol_number-1] = ol_amount;
                total_amount += ol_amount;

                if(i_data.indexOf("generic") != -1 && s_data.indexOf("generic") != -1)
                {
                    brandGeneric[ol_number-1] = 'B';
                }
                else
                {
                    brandGeneric[ol_number-1] = 'G';
                }

                switch((int)d_id)
                {
                    case 1: ol_dist_info = s_dist_01; break;
                    case 2: ol_dist_info = s_dist_02; break;
                    case 3: ol_dist_info = s_dist_03; break;
                    case 4: ol_dist_info = s_dist_04; break;
                    case 5: ol_dist_info = s_dist_05; break;
                    case 6: ol_dist_info = s_dist_06; break;
                    case 7: ol_dist_info = s_dist_07; break;
                    case 8: ol_dist_info = s_dist_08; break;
                    case 9: ol_dist_info = s_dist_09; break;
                    case 10: ol_dist_info = s_dist_10; break;
                }

                query = new StringBuffer();
                query.append("INSERT INTO order_line (ol_o_id, ol_d_id, ol_w_id, ol_number,");
                query.append(" ol_i_id, ol_supply_w_id,");
                query.append(" ol_quantity, ol_amount, ol_dist_info, ol_delivery_d)");
                query.append(" VALUES (");
                query.append(o_id);
                query.append(", ");
                query.append(d_id);
                query.append(", ");
                query.append(w_id);
                query.append(", ");
                query.append(ol_number);
                query.append(", ");
                query.append(ol_i_id);
                query.append(", ");
                query.append(ol_supply_w_id);
                query.append(", ");
                query.append(ol_quantity);
                query.append(", ");
                query.append(ol_amount);
                query.append(", \'");
                query.append(ol_dist_info);
                query.append("\', NULL)");
                printMessage(query.toString());
                stmt.executeUpdate(query.toString());
            }
            total_amount *= (1+w_tax+d_tax)*(1-c_discount);


            conn.commit();


            StringBuffer terminalMessage = new StringBuffer();
            terminalMessage.append("\n+--------------------------- NEW-ORDER ---------------------------+\n");
            terminalMessage.append(" Date: ");
            terminalMessage.append(getCurrentTime());
            terminalMessage.append("\n\n Warehouse: ");
            terminalMessage.append(w_id);
            terminalMessage.append("\n   Tax:     ");
            terminalMessage.append(w_tax);
            terminalMessage.append("\n District:  ");
            terminalMessage.append(d_id);
            terminalMessage.append("\n   Tax:     ");
            terminalMessage.append(d_tax);
            terminalMessage.append("\n Order:     ");
            terminalMessage.append(o_id);
            terminalMessage.append("\n   Lines:   ");
            terminalMessage.append(o_ol_cnt);
            terminalMessage.append("\n\n Customer:  ");
            terminalMessage.append(c_id);
            terminalMessage.append("\n   Name:    ");
            terminalMessage.append(c_last);
            terminalMessage.append("\n   Credit:  ");
            terminalMessage.append(c_credit);
            terminalMessage.append("\n   %Disc:   ");
            terminalMessage.append(c_discount);
            terminalMessage.append("\n\n Order-Line List [Supp_W - Item_ID - Item Name - Qty - Stock - B/G - Price - Amount]\n");
            for(int i = 0; i < o_ol_cnt; i++)
            {
                terminalMessage.append("                 [");
                terminalMessage.append(supplierWarehouseIDs[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(itemIDs[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(itemNames[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(orderQuantities[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(stockQuantities[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(brandGeneric[i]);
                terminalMessage.append(" - ");
                terminalMessage.append(formattedDouble(itemPrices[i]));
                terminalMessage.append(" - ");
                terminalMessage.append(formattedDouble(orderLineAmounts[i]));
                terminalMessage.append("]\n");
            }
            terminalMessage.append("\n\n Total Amount: ");
            terminalMessage.append(total_amount);
            terminalMessage.append("\n\n Execution Status: New order placed!\n");
            terminalMessage.append("+-----------------------------------------------------------------+\n\n");
            terminalMessage(terminalMessage.toString());
        }
        catch(Exception e)
        {
            if(e instanceof IllegalAccessException)
            {
                StringBuffer terminalMessage = new StringBuffer();
                terminalMessage.append("\n+--------------------------- NEW-ORDER ---------------------------+");
                terminalMessage.append("\n Warehouse: ");
                terminalMessage.append(w_id);
                terminalMessage.append("\n District:  ");
                terminalMessage.append(d_id);
                terminalMessage.append("\n Order:     ");
                terminalMessage.append(o_id);
                terminalMessage.append("\n\n Customer:  ");
                terminalMessage.append(c_id);
                terminalMessage.append("\n   Name:    ");
                terminalMessage.append(c_last);
                terminalMessage.append("\n   Credit:  ");
                terminalMessage.append(c_credit);
                terminalMessage.append("\n\n Execution Status: Item number is not valid!\n");
                terminalMessage.append("+-----------------------------------------------------------------+\n\n");
                terminalMessage(terminalMessage.toString());
            }
            else
            {
                error("NEW-ORDER");
                logException(e);
            }

            try
            {
                terminalMessage("Performing ROLLBACK...");
                conn.rollback();
            }
            catch(Exception e1)
            {
                error("NEW-ORDER-ROLLBACK");
                logException(e1);
            }
        }
    }


    private void stockLevelTransaction(long w_id, long d_id, long threshold)
    {
        long o_id, stock_count;

        try
        {
            query = new StringBuffer();
            query.append("SELECT d_next_o_id");
            query.append(" FROM district");
            query.append(" WHERE d_w_id = ");
            query.append(w_id);
            query.append(" AND d_id = ");
            query.append(d_id);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(!rs.next()) throw new Exception("D_W_ID=" + w_id + " D_ID=" + d_id + " not found!");
            o_id = rs.getLong("d_next_o_id");
            rs.close();
            rs = null;


            query = new StringBuffer();
            query.append("SELECT COUNT(DISTINCT (s_i_id)) AS stock_count");
            query.append(" FROM order_line, stock");
            query.append(" WHERE ol_w_id = ");
            query.append(w_id);
            query.append(" AND ol_d_id = ");
            query.append(d_id);
            query.append(" AND ol_o_id < ");
            query.append(o_id);
            query.append(" AND ol_o_id >= ");
            query.append(o_id-20);
            query.append(" AND s_w_id = ");
            query.append(w_id);
            query.append(" AND s_i_id = ol_i_id");
            query.append(" AND s_quantity < ");
            query.append(threshold);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(!rs.next()) throw new Exception("OL_W_ID=" + w_id + " OL_D_ID=" + d_id + " OL_O_ID=" + o_id + " (...) not found!");
            stock_count = rs.getLong("stock_count");
            rs.close();
            rs = null;


            StringBuffer terminalMessage = new StringBuffer();
            terminalMessage.append("\n+-------------------------- STOCK-LEVEL --------------------------+");
            terminalMessage.append("\n Warehouse: ");
            terminalMessage.append(w_id);
            terminalMessage.append("\n District:  ");
            terminalMessage.append(d_id);
            terminalMessage.append("\n\n Stock Level Threshold: ");
            terminalMessage.append(threshold);
            terminalMessage.append("\n Low Stock Count:       ");
            terminalMessage.append(stock_count);
            terminalMessage.append("\n+-----------------------------------------------------------------+\n\n");
            terminalMessage(terminalMessage.toString());
        }
        catch(Exception e)
        {
            error("STOCK-LEVEL");
            logException(e);
        }
    }

    private void paymentTransaction(long w_id, long c_w_id, double h_amount, long d_id, long c_d_id, long c_id, String c_last, boolean c_by_name)
    {
        String w_street_1, w_street_2, w_city, w_state, w_zip, w_name;
        String d_street_1, d_street_2, d_city, d_state, d_zip, d_name;
        long namecnt;
        String c_first, c_middle, c_street_1, c_street_2, c_city, c_state, c_zip;
        String c_phone, c_credit = null, c_data = null, c_new_data, h_data;
        double c_credit_lim, c_discount, c_balance = 0;
        java.sql.Date c_since;


        try
        {
            query = new StringBuffer();
            query.append("UPDATE warehouse SET w_ytd = w_ytd + ");
            query.append(h_amount);
            query.append(" WHERE w_id = ");
            query.append(w_id);
            printMessage(query.toString());
            result = stmt.executeUpdate(query.toString());
            if(result == 0) throw new Exception("W_ID=" + w_id + " not found!");


            query = new StringBuffer();
            query.append("SELECT w_street_1, w_street_2, w_city, w_state, w_zip, w_name");
            query.append(" FROM warehouse");
            query.append(" WHERE w_id = ");
            query.append(w_id);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(!rs.next()) throw new Exception("W_ID=" + w_id + " not found!");
            w_street_1 = rs.getString("w_street_1");
            w_street_2 = rs.getString("w_street_2");
            w_city = rs.getString("w_city");
            w_state = rs.getString("w_state");
            w_zip = rs.getString("w_zip");
            w_name = rs.getString("w_name");
            rs.close();
            rs = null;


            query = new StringBuffer();
            query.append("UPDATE district SET d_ytd = d_ytd + ");
            query.append(h_amount);
            query.append(" WHERE d_w_id = ");
            query.append(w_id);
            query.append(" AND d_id = ");
            query.append(d_id);
            printMessage(query.toString());
            result = stmt.executeUpdate(query.toString());
            if(result == 0) throw new Exception("D_ID=" + d_id + " D_W_ID=" + w_id + " not found!");


            query = new StringBuffer();
            query.append("SELECT d_street_1, d_street_2, d_city, d_state, d_zip, d_name");
            query.append(" FROM district");
            query.append(" WHERE d_w_id = ");
            query.append(w_id);
            query.append(" AND d_id = ");
            query.append(d_id);
            printMessage(query.toString());
            rs = stmt.executeQuery(query.toString());
            if(!rs.next()) throw new Exception("D_ID=" + d_id + " D_W_ID=" + w_id + " not found!");
            d_street_1 = rs.getString("d_street_1");
            d_street_2 = rs.getString("d_street_2");
            d_city = rs.getString("d_city");
            d_state = rs.getString("d_state");
            d_zip = rs.getString("d_zip");
            d_name = rs.getString("d_name");
            rs.close();
            rs = null;


            if(c_by_name)
            {
                query = new StringBuffer();
                query.append("SELECT count(c_id) AS namecnt");
                query.append(" FROM customer");
                query.append(" WHERE c_last = \'");
                query.append(c_last);
                query.append("\' AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_w_id = ");
                query.append(c_w_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_LAST=" + c_last + " C_D_ID=" + c_d_id + " C_W_ID=" + c_w_id + " not found!");
                namecnt = rs.getLong("namecnt");
                rs.close();
                rs = null;


                query = new StringBuffer();
                query.append("SELECT c_first, c_middle, c_id,");
                query.append(" c_street_1, c_street_2, c_city, c_state, c_zip,");
                query.append(" c_phone, c_credit, c_credit_lim,");
                query.append(" c_discount, c_balance, c_since");
                query.append(" FROM customer");
                query.append(" WHERE c_w_id = ");
                query.append(c_w_id);
                query.append(" AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_last = \'");
                query.append(c_last);
                query.append("\' ORDER BY c_first ASC");
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_LAST=" + c_last + " C_D_ID=" + c_d_id + " C_W_ID=" + c_w_id + " not found!");
                if(namecnt%2 == 1) namecnt++;
                for(int i = 1; i < namecnt / 2; i++) rs.next();
                c_id = rs.getLong("c_id");
                c_first = rs.getString("c_first");
                c_middle = rs.getString("c_middle");
                c_street_1 = rs.getString("c_street_1");
                c_street_2 = rs.getString("c_street_2");
                c_city = rs.getString("c_city");
                c_state = rs.getString("c_state");
                c_zip = rs.getString("c_zip");
                c_phone = rs.getString("c_phone");
                c_credit = rs.getString("c_credit");
                c_credit_lim = rs.getDouble("c_credit_lim");
                c_discount = rs.getDouble("c_discount");
                c_balance = rs.getDouble("c_balance");
                c_since = rs.getDate("c_since");
                rs.close();
                rs = null;
            }
            else
            {
                query = new StringBuffer();
                query.append("SELECT c_first, c_middle, c_last,");
                query.append(" c_street_1, c_street_2, c_city, c_state, c_zip,");
                query.append(" c_phone, c_credit, c_credit_lim,");
                query.append(" c_discount, c_balance, c_since");
                query.append(" FROM customer");
                query.append(" WHERE c_w_id = ");
                query.append(c_w_id);
                query.append(" AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_id = ");
                query.append(c_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_ID=" + c_id + " C_D_ID=" + c_d_id + " C_W_ID=" + c_w_id + " not found!");
                c_last = rs.getString("c_last");
                c_first = rs.getString("c_first");
                c_middle = rs.getString("c_middle");
                c_street_1 = rs.getString("c_street_1");
                c_street_2 = rs.getString("c_street_2");
                c_city = rs.getString("c_city");
                c_state = rs.getString("c_state");
                c_zip = rs.getString("c_zip");
                c_phone = rs.getString("c_phone");
                c_credit = rs.getString("c_credit");
                c_credit_lim = rs.getDouble("c_credit_lim");
                c_discount = rs.getDouble("c_discount");
                c_balance = rs.getDouble("c_balance");
                c_since = rs.getDate("c_since");
                rs.close();
                rs = null;
            }


            c_balance += h_amount;
            if(c_credit.equals("BC"))
            {
                query = new StringBuffer();
                query.append("SELECT c_data");
                query.append(" FROM customer");
                query.append(" WHERE c_w_id = ");
                query.append(c_w_id);
                query.append(" AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_id = ");
                query.append(c_id);
                printMessage(query.toString());
                rs = stmt.executeQuery(query.toString());
                if(!rs.next()) throw new Exception("C_ID=" + c_id + " C_W_ID=" + c_w_id + " C_D_ID=" + c_d_id + " not found!");
                c_data = rs.getString("c_data");
                rs.close();
                rs = null;


                c_new_data = c_id + " " + c_d_id + " " + c_w_id + " " + d_id + " " + w_id  + " " + h_amount + " |";
                if(c_data.length() > c_new_data.length())
                {
                    c_new_data += c_data.substring(0, c_data.length()-c_new_data.length());
                }
                else
                {
                    c_new_data += c_data;
                }
                if(c_new_data.length() > 500) c_new_data = c_new_data.substring(0, 500);
                query = new StringBuffer();
                query.append("UPDATE customer SET c_balance = ");
                query.append(c_balance);
                query.append(", c_data = \'");
                query.append(c_new_data);
                query.append("\' WHERE c_w_id = ");
                query.append(c_w_id);
                query.append(" AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_id = ");
                query.append(c_id);
                printMessage(query.toString());
                result = stmt.executeUpdate(query.toString());
                if(result == 0) throw new Exception("C_ID=" + c_id + " C_W_ID=" + c_w_id + " C_D_ID=" + c_d_id + " not found!");
            }
            else
            {
                query = new StringBuffer();
                query.append("UPDATE customer SET c_balance = ");
                query.append(c_balance);
                query.append(" WHERE c_w_id = ");
                query.append(c_w_id);
                query.append(" AND c_d_id = ");
                query.append(c_d_id);
                query.append(" AND c_id = ");
                query.append(c_id);
                printMessage(query.toString());
                result = stmt.executeUpdate(query.toString());
                if(result == 0) throw new Exception("C_ID=" + c_id + " C_W_ID=" + c_w_id + " C_D_ID=" + c_d_id + " not found!");
            }


            if(w_name.length() > 10) w_name = w_name.substring(0, 10);
            if(d_name.length() > 10) d_name = d_name.substring(0, 10);
            h_data = w_name + "    " + d_name;
            query = new StringBuffer();
            query.append("INSERT INTO history (h_c_d_id, h_c_w_id, h_c_id, h_d_id,");
            query.append(" h_w_id, h_date, h_amount, h_data)");
            query.append(" VALUES (");
            query.append(c_d_id);
            query.append(", ");
            query.append(c_w_id);
            query.append(", ");
            query.append(c_id);
            query.append(", ");
            query.append(d_id);
            query.append(", ");
            query.append(w_id);
            query.append(", SYSDATE");
            query.append(", ");
            query.append(h_amount);
            query.append(", \'");
            query.append(h_data);
            query.append("\')");
            printMessage(query.toString());
            stmt.executeUpdate(query.toString());


            conn.commit();


            StringBuffer terminalMessage = new StringBuffer();
            terminalMessage.append("\n+---------------------------- PAYMENT ----------------------------+");
            terminalMessage.append("\n Date: " + getCurrentTime());
            terminalMessage.append("\n\n Warehouse: ");
            terminalMessage.append(w_id);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(w_street_1);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(w_street_2);
            terminalMessage.append("\n   City:    ");
            terminalMessage.append(w_city);
            terminalMessage.append("   State: ");
            terminalMessage.append(w_state);
            terminalMessage.append("  Zip: ");
            terminalMessage.append(w_zip);
            terminalMessage.append("\n\n District:  ");
            terminalMessage.append(d_id);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(d_street_1);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(d_street_2);
            terminalMessage.append("\n   City:    ");
            terminalMessage.append(d_city);
            terminalMessage.append("   State: ");
            terminalMessage.append(d_state);
            terminalMessage.append("  Zip: ");
            terminalMessage.append(d_zip);
            terminalMessage.append("\n\n Customer:  ");
            terminalMessage.append(c_id);
            terminalMessage.append("\n   Name:    ");
            terminalMessage.append(c_first);
            terminalMessage.append(" ");
            terminalMessage.append(c_middle);
            terminalMessage.append(" ");
            terminalMessage.append(c_last);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(c_street_1);
            terminalMessage.append("\n   Street:  ");
            terminalMessage.append(c_street_2);
            terminalMessage.append("\n   City:    ");
            terminalMessage.append(c_city);
            terminalMessage.append("   State: ");
            terminalMessage.append(c_state);
            terminalMessage.append("  Zip: ");
            terminalMessage.append(c_zip);
            terminalMessage.append("\n   Since:   ");
            terminalMessage.append(c_since.toString());
            terminalMessage.append("\n   Credit:  ");
            terminalMessage.append(c_credit);
            terminalMessage.append("\n   %Disc:   ");
            terminalMessage.append(c_discount);
            terminalMessage.append("\n   Phone:   ");
            terminalMessage.append(c_phone);
            terminalMessage.append("\n\n Amount Paid:      ");
            terminalMessage.append(h_amount);
            terminalMessage.append("\n Credit Limit:     ");
            terminalMessage.append(c_credit_lim);
            terminalMessage.append("\n New Cust-Balance: ");
            terminalMessage.append(c_balance);
            if(c_credit.equals("BC"))
            {
                if(c_data.length() > 40)
                {
                    terminalMessage.append("\n\n Cust-Data: " + c_data.substring(0, 50));
                    int data_chunks = c_data.length() > 200 ? 4 : c_data.length()/50;
                    for(int n = 1; n < data_chunks; n++)
                        terminalMessage.append("\n            " + c_data.substring(n*50, (n+1)*50));
                }
                else
                {
                    terminalMessage.append("\n\n Cust-Data: " + c_data);
                }
            }
            terminalMessage.append("\n+-----------------------------------------------------------------+\n\n");
            terminalMessage(terminalMessage.toString());
        }
        catch(Exception e)
        {
            error("PAYMENT");
            logException(e);
            try
            {
                terminalMessage("Performing ROLLBACK...");
                conn.rollback();
            }
            catch(Exception e1)
            {
                error("PAYMENT-ROLLBACK");
                logException(e1);
            }
        }
    }

    private String lastName(int num)
    {
        return nameTokens[num/100] + nameTokens[(num/10)%10] + nameTokens[num%10];
    }

    private long randomNumber(long min, long max)
    {
        return (long)(random.nextDouble() * (max-min+1) + min);
    }

    private long nonUniformRandom(long x, long min, long max)
    {
        return (((randomNumber(0, x) | randomNumber(min, max)) + randomNumber(0, x)) % (max-min+1)) + min;
    }

    private void error(String type)
    {
        errorOutputArea.println("[ERROR] TERMINAL=" + terminalName + "  TYPE=" + type + "  COUNT=" + transactionCount);
    }

    private void logException(Exception e)
    {
        StringWriter stringWriter = new StringWriter();
        PrintWriter printWriter = new PrintWriter(stringWriter);
        e.printStackTrace(printWriter);
        printWriter.close();
        errorOutputArea.println(stringWriter.toString());
    }

    private void terminalMessage(String message)
    {
        if(TERMINAL_MESSAGES) terminalOutputArea.println(message);
    }

    private void printMessage(String message)
    {
        if(debugMessages) terminalOutputArea.println("[ jTPCC ] " + message);
    }

    private String getCurrentTime()
    {
        return dateFormat.format(new java.util.Date());
    }

    private String formattedDouble(double d)
    {
        String dS = ""+d;
        return dS.length() > 6 ? dS.substring(0, 6) : dS;
    }
}